/*
 * This represents a reader thread in the simulator which keeps on reading from EMUTHREADS.
 */

package emulatorinterface;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Array;
import java.util.ArrayList;

import main.Main;
import pipeline.PipelineInterface;
import config.SystemConfig;
import emulatorinterface.communication.Encoding;
import emulatorinterface.communication.IpcBase;
import emulatorinterface.communication.Packet;
import emulatorinterface.communication.shm.SharedMem;
import emulatorinterface.translator.qemuTranslationCache.TranslatedInstructionCache;
import emulatorinterface.translator.visaHandler.VisaHandlerSelector;
import emulatorinterface.translator.x86.objparser.ObjParser;
import generic.CircularPacketQueue;
import generic.ClockDomain;
import generic.Core;
import generic.GenericCircularQueue;
import generic.Instruction;
import generic.OperationType;
import generic.Statistics;

/* MaxNumThreads threads are created from this class. Each thread
 * continuously keeps reading from the shared memory segment according
 * to its index(taken care in the jni C file).
 */
public class RunnableThread implements Encoding, Runnable {

	int tid;
	int EMUTHREADS = SystemConfig.numEmuThreadsPerJavaThread;
	int currentEMUTHREADS = 0;

	public IpcBase ipcType;

	ClockDomain clockDomain;

	EmulatorThreadState[] threadParams = new EmulatorThreadState[EMUTHREADS];
	boolean[] error;

	GenericCircularQueue<Instruction>[] inputToPipeline;

	DynamicInstructionBuffer[] dynamicInstructionBuffer;

	long[] noOfMicroOps;
	long[] numInstructions;
	PipelineInterface[] pipelineInterfaces;
	boolean backoff = false;
	public VisaHandlerSelector myVisa;
	public TranslatedInstructionCache myTranCache;
	private long start = 0;
	PrintWriter writer;
	long totalPackets = 0;

	// initialise a reader thread with the correct thread id and the buffer to
	// write the results in.
	@SuppressWarnings("unchecked")
	public RunnableThread(String threadName, int tid1, IpcBase ipcType, Core[] cores) throws FileNotFoundException,
			UnsupportedEncodingException {
		String logFile = "/mnt/srishtistr0/home/geetika/outOfOrderWorkspace/ParTejas/log_con_" + tid1;
		writer = new PrintWriter(logFile, "UTF-8");
		// System.out.println("seep my name is "+threadName);
		this.ipcType = ipcType;

		dynamicInstructionBuffer = new DynamicInstructionBuffer[EMUTHREADS];
		inputToPipeline = (GenericCircularQueue<Instruction>[]) Array.newInstance(GenericCircularQueue.class,
				EMUTHREADS);

		myVisa = new VisaHandlerSelector();
		myTranCache = new TranslatedInstructionCache();
		// initialize the pool
		noOfMicroOps = new long[EMUTHREADS];
		numInstructions = new long[EMUTHREADS];
		pipelineInterfaces = new PipelineInterface[EMUTHREADS];
		error = new boolean[EMUTHREADS];
		for (int tidEmu = 0; tidEmu < EMUTHREADS; tidEmu++) {
			int tidApp = IpcBase.getAppId(tid1, tidEmu);// tid1*IpcBase.EmuThreadsPerJavaThread+i;
			IpcBase.glTable.getStateTable().put(tidApp, new ThreadState(tidApp));
			threadParams[tidEmu] = new EmulatorThreadState();
			error[tidEmu] = true;
			// TODO pipelineinterfaces & inputToPipeline should also be in the
			// IpcBase
			pipelineInterfaces[tidEmu] = cores[tidApp].getPipelineInterface();
			inputToPipeline[tidEmu] = new GenericCircularQueue<Instruction>(Instruction.class,
					Main.INSTRUCTION_THRESHOLD);

			dynamicInstructionBuffer[tidEmu] = new DynamicInstructionBuffer();
			GenericCircularQueue<Instruction>[] toBeSet = (GenericCircularQueue<Instruction>[]) Array.newInstance(
					GenericCircularQueue.class, 1);
			toBeSet[0] = inputToPipeline[tidEmu];
			pipelineInterfaces[tidEmu].setInputToPipeline(toBeSet);
		}
		this.tid = tid1;
		System.out.println("--  starting java thread" + this.tid);
	}

	public void run() {

		// System.out.println("-- in runnable thread run "+this.tid);
		ArrayList<CircularPacketQueue> fromEmulatorAll = new ArrayList<CircularPacketQueue>(EMUTHREADS);

		for (int i = 0; i < EMUTHREADS; i++) {
			CircularPacketQueue fromEmulator = new CircularPacketQueue(SharedMem.COUNT);
			fromEmulatorAll.add(fromEmulator);
		}

		Packet pnew = new Packet();
		boolean allover = false;
		boolean emulatorStarted = false;

		long totalPackets[] = new long[EMUTHREADS];
		long junkSum[] = new long[EMUTHREADS];

		EmulatorThreadState thread;
		// keep on looping till there is something to read. iterates on the
		// emulator threads from which it has to read.
		// tid is java thread id
		// tidEmu is the local notion of pin threads for the current java thread
		// tidApp is the actual tid of a pin thread
		int tidEmu;
		while (true) {
			for (tidEmu = 0; tidEmu < EMUTHREADS; tidEmu++) {
				shouldBackoff(emulatorStarted);

				CircularPacketQueue fromEmulator = fromEmulatorAll.get(tidEmu);

				thread = threadParams[tidEmu];
				if (thread.halted || thread.finished) {
					continue;
				}

				if (thread.outstandingMicroOps.isEmpty() == false) {
					// System.out.println(javaTid + ":  RT5");
					if (thread.outstandingMicroOps.size() < inputToPipeline[tidEmu].spaceLeft()) {
						while (thread.outstandingMicroOps.isEmpty() == false) {
							inputToPipeline[tidEmu].enqueue(thread.outstandingMicroOps.pollFirst());
						}
					} else {
						// System.out.println(javaTid + ":  RT6");
						// there is no space in pipelineBuffer. So don't fetch
						// any more instructions
						continue;
					}
				}

				int tidApp = IpcBase.getAppId(tid, tidEmu);// tid * EMUTHREADS +
															// tidEmu;
				int numReads = 0;
				long v = 0;

				// if we read -1, this means this emulator thread finished.
				/*
				 * if (error[tidEmu] == false &&
				 * inputToPipeline[tidEmu].isEmpty() &&
				 * pipelineInterfaces[tidEmu
				 * ].getCore().getEventQueue().getPriorityQueue().isEmpty()) {
				 * Main.finished[tidApp] = true; // System.out.println("THREAD "
				 * + tidApp + // " ------- Finished ----  "); if
				 * (Main.allFinished()) { signalFinish(tidEmu); } }
				 */

				// If java thread itself is terminated then break out from this
				// for loop. also update the variable allover so that I can
				// break from$the while loop also.
				if (ipcType.termination[tid] == true) {
					allover = true;
					break;
				}

				numReads = ipcType.fetchManyPackets(tidApp, fromEmulator);

				if (fromEmulator.size()/* numReads */== 0) {
					continue;
				}

				// need to do this only the first time
				if (!emulatorStarted) {
					emulatorStarted = true;
					iStarted();
				}

				thread.checkStarted();

				thread.totalRead += numReads;

				/*
				 * if(numReads > 0) for(long j = 0; j < 1000000000l; j +=
				 * numReads) { junkSum[tidEmu] += j; }
				 */

				for (int i = 0; i < /* numReads */fromEmulator.size(); i++) {
					pnew = fromEmulator.dequeue();
					v = pnew.value;
//					if (pnew.value == Encoding.ASSEMBLY) {
//						byte asmBytes[] = Main.myPool[tidApp].getCustomAsmCharPool().dequeue(tidApp);
//					}

					totalPackets[tidEmu]++;
					if (totalPackets[tidEmu] % 10000000 == 0) {
						writer.println(tidEmu + "total packets till now " + totalPackets[tidEmu] + " "
								+ junkSum[tidEmu]);
						junkSum[tidEmu] = 0;
						writer.flush();
					}
					/*
					 * if(totalPackets[tidEmu]%10000 == 0)
					 * System.out.println(tidEmu + "total packets till now " +
					 * totalPackets[tidEmu]);
					 * //System.out.println(" Reading packet " + tidApp +"   " +
					 * i);
					 */
					if (v == Encoding.THREADCOMPLETE) {
						signalFinish(tidEmu);
						// pipelineInterfaces[tidEmu].setExecutionComplete(true);

					}

					/*
					 * boolean done = true; for(int j = 0; j < EMUTHREADS; j++)
					 * { if(j == 0 && totalPackets[j] < 10) { done = false; }
					 * if(j != 0 && totalPackets[j] < 1000000) { done = false; }
					 * } if(done) { System.out.println(tidApp +
					 * " Signal Finish "); signalFinish(tidEmu); }
					 */
					if (v == Encoding.SUBSETSIMCOMPLETE) {
						boolean loopBreak = false;
						for (int javaTid = 0; javaTid < SystemConfig.maxNumJavaThreads; javaTid++) {
							for (int emuTid = 0; emuTid < SystemConfig.numEmuThreadsPerJavaThread; emuTid++) {
								if (Main.runners[javaTid].inputToPipeline[emuTid].isFull()) {
									// This is being done because we cannot
									// directly enqueue the invalid
									// packet to all the pipelines, some of the
									// pipelines might not have
									// enough space in their input to pipeline
									loopBreak = true;
									fromEmulator.enqueue(pnew.ip, pnew.value, pnew.tgt);
									break;
								}
								Main.runners[javaTid].inputToPipeline[emuTid].enqueue(new Instruction(
										OperationType.inValid, null, null, null));
								int globalTid = IpcBase.getAppId(javaTid, emuTid);
								IpcBase.glTable.getStateTable().get((Integer) globalTid).lastTimerseen = Long.MAX_VALUE;// (long)-1>>>1;
								// System.out.println(tidApp+" pin thread got -1");

								threadParams[emuTid].finished = true;

								// hack, should ideally listen for fini
								// function from pin. although this
								// will always work
								if (globalTid == 0) {
									for (int k = 0; k < SystemConfig.maxNumJavaThreads; k++)
										ipcType.termination[k] = true;
								}
							}
							if (loopBreak)
								break;
						}
						if (loopBreak)
							break;
					}

					// System.out.println(tid + "    " + pnew.toString());
					boolean ret = processPacket(thread, pnew, tidEmu);
					if (ret == false) {
						// There is not enough space in pipeline buffer.
						// So don't process any more packets.
						break;
					}
				}

				boolean myerror = ipcType.errorCheck(tidApp, numReads);
				if (myerror == false) {
					// error[tidEmu] = false;
					System.out.println(tidApp + "  got the file end packet  ");
					signalFinish(tidEmu);
				}

				if (ipcType.termination[tid] == true) {
					allover = true;
					break;
				}
			}
			// System.out.println(tid + ": RS1");
			runPipelines();
			// System.out.println(tid + ": RS2");
			// System.out.println("after execution n=  "+n+" Thread finished ? "+threadParams[1].finished);

			// this runnable thread can be stopped in two ways. Either the
			// emulator threads from which it was supposed to read never
			// started(none of them) so it has to be signalled by the main
			// thread. When this happens 'allover' becomes 'true' and it
			// breaks out from the loop. The second situation is that all the
			// emulator threads which started have now finished, so probably
			// this thread should now terminate.
			// The second condition handles this situation.
			// NOTE this ugly state management cannot be avoided unless we use
			// some kind of a signalling mechanism between the emulator and
			// simulator(TODO).
			// Although this should handle most of the cases.
			if (allover || (emulatorStarted && emuThreadsFinished())) {
				ipcType.termination[tid] = true;
				break;
			}
		}
		finishAllPipelines();
	}

	// Todo implement a backoff scheme here.
	private void shouldBackoff(boolean emulatorStarted) {
		if (emulatorStarted)
			return;
		if (!backoff) {
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			backoff = true;
		} else
			backoff = false;
	}

	public void setClockDomain(ClockDomain clockDomain) {
		this.clockDomain = clockDomain;
	}

	protected ClockDomain getClockDomain() {
		return this.clockDomain;
	}

	// This function must be called only once o/w would cause havoc
	protected void iStarted() {
		Main.barrierClocks.register();
		Main.barrierClocks.arriveAndAwaitAdvance();
		System.out.println("tid" + this.tid + "Registered" + Main.barrierClocks.getRegisteredParties());

		this.start = System.currentTimeMillis();
		Main.startTime.compareAndSet(-1, this.start);
		ipcType.started[tid] = true;
	}

	void iFinished() {

		if (ipcType.started[tid]) {
			if (this.clockDomain.previousPhase == Main.barrierClocks.getPhase()) {
				Main.barrierClocks.awaitAdvance(this.clockDomain.previousPhase);
			}
			Main.barrierClocks.arriveAndDeregister();

		}

		long dataRead = 0;
		long totNumIns = 0;
		long totMicroOps = 0;
		for (int i = 0; i < EMUTHREADS; i++) {
			totMicroOps += noOfMicroOps[i];
			dataRead += threadParams[i].totalRead;
			totNumIns += numInstructions[i];
		}
		long timeTaken = System.currentTimeMillis() - this.start;
		if (dataRead != 0)
			System.out.println("\nThread" + tid + " Bytes-" + dataRead * 24 + " instructions-" + totNumIns
					+ " microOps  " + totMicroOps + " MBPS-" + (double) (dataRead * 24) / (double) timeTaken / 1000.0
					+ " time-" + timeTaken + "\n microOp KIPS- " + (double) totMicroOps / (double) timeTaken + " KIPS-"
					+ (double) totNumIns / (double) timeTaken);

		Statistics.setDataRead(dataRead * 24, tid);
		Statistics.setNoOfMicroOps(noOfMicroOps, tid);

		IpcBase.free.release();

	}

	protected void runPipelines() {
		int minN = Integer.MAX_VALUE;
		int n[] = new int[EMUTHREADS];
		for (int tidEmu = 0; tidEmu < currentEMUTHREADS; tidEmu++) {
			EmulatorThreadState th = threadParams[tidEmu];
			if (th.halted && !(this.inputToPipeline[tidEmu].size() > Main.INSTRUCTION_THRESHOLD)) {
				// || th.packets.size() > PACKET_THRESHOLD)){
				th.halted = false;
				// System.out.println("Halting over..!! "+IpcBase.getAppId(tid,
				// tidEmu));
			}
			// System.out.println("inputtopipeline size of thread "+tidEmu+
			// " is "+ inputToPipeline[tidEmu].getListSize());
			n[tidEmu] = (inputToPipeline[tidEmu].size() / pipelineInterfaces[tidEmu].getCore().getDecodeWidth())
					* pipelineInterfaces[tidEmu].getCoreStepSize();
			if (n[tidEmu] < minN && n[tidEmu] != 0)
				minN = n[tidEmu];
		}

		// 10 empty cycles if nothing to proceed
		if (minN == Integer.MAX_VALUE) {
			if (ipcType.started[tid]) {
				minN = 10;
				for (int tidEmu = 0; tidEmu < EMUTHREADS; tidEmu++) {
					Statistics.emptyCycles[IpcBase.getAppId(tid, tidEmu)] += minN;
				}
			} else
				minN = 0;
		}

		for (int i1 = 0; i1 < minN; i1++) {
			// //System.out.println(tid + ": RP1");
			for (int tidEmu = 0; tidEmu < currentEMUTHREADS; tidEmu++) {
				pipelineInterfaces[tidEmu].oneCycleOperation();
			}
			// //System.out.println(tid + ": RP2");
			// System.out.println("seep one cycle operation  "+
			// clockDomain.getCurrentTime());
			clockDomain.incrementClock();
			// //System.out.println(tid + ": RP3");
		}

	}

	public void finishAllPipelines() {
		for (int i = 0; i < currentEMUTHREADS; i++) {
			while (this.inputToPipeline[i].isFull())
				runPipelines();
			this.inputToPipeline[i].enqueue(new Instruction(OperationType.inValid, null, null, null));
		}
		for (int i = 0; i < currentEMUTHREADS; i++) {
			if (!pipelineInterfaces[i].isExecutionComplete() && pipelineInterfaces[i].isSleeping()) {
				System.out.println("not completed for " + i);
				// resumeSleep(IpcBase.glTable.resumePipelineTimer(i));
				this.pipelineInterfaces[i].resumePipeline();
			}
		}

		boolean queueComplete; // queueComplete is true when all cores have
								// completed
		// int counter = 0;
		while (true) {
			// counter++;
			// System.out.println("Pin completed ");
			queueComplete = true;
			for (int i = 0; i < currentEMUTHREADS; i++) {
				queueComplete = queueComplete && pipelineInterfaces[i].isExecutionComplete();
			}
			if (queueComplete == true) {
				break;
			}

			for (int tidEmu = 0; tidEmu < currentEMUTHREADS; tidEmu++) {
				pipelineInterfaces[tidEmu].oneCycleOperation();
			}
			clockDomain.incrementClock();
		}
		for (int i = 0; i < EMUTHREADS; i++)
			Main.finished[IpcBase.getAppId(tid, i)] = true;
		// System.out.println(tid + "  11111 ");
		while (!Main.allFinished()) {
			for (int tidEmu = 0; tidEmu < currentEMUTHREADS; tidEmu++) {
				pipelineInterfaces[tidEmu].drainEventQueue();
			}
			clockDomain.incrementClock();
		}
		long end = System.currentTimeMillis();
		Main.endTime.set(end);
		// System.out.println(tid + "  time :  " + (end -
		// Main.startTime.get()));
		iFinished();
	}

	// returns true if all the emulator threads from which I was reading have
	// finished
	protected boolean emuThreadsFinished() {
		boolean ret = true;
		for (int i = 0; i < EMUTHREADS; i++) {
			EmulatorThreadState thread = threadParams[i];

			// Fixed: the thread is not finished as new threads may get spawned
			if (!thread.started)
				return false;

			if (thread.started == true && thread.finished == false) {
				return false;
			}
		}
		return ret;
	}

	protected boolean processPacket(EmulatorThreadState thread, Packet pnew, int tidEmu) {

		boolean isSpaceInPipelineBuffer = true;
		int tidApp = IpcBase.getAppId(this.tid, tidEmu);

		if (pnew.value == TIMER) {
			// resumeSleep(IpcBase.glTable.tryResumeOnWaitingPipelines(tidApp,
			// pnew.ip));
			return true;
		}
		if (pnew.value > SYNCHSTART && pnew.value < SYNCHEND) {
			// resumeSleep(IpcBase.glTable.update(pnew.tgt, tidApp, pnew.ip,
			// pnew.value));
			return true;
		}
		if (thread.isFirstPacket) {
			currentEMUTHREADS++;
			thread.pold = pnew;
			thread.isFirstPacket = false;
			pipelineInterfaces[tidEmu].getCore().getExecEngine().setExecutionComplete(false);
			pipelineInterfaces[tidEmu].getCore().getExecEngine().setExecutionBegun(true);
			this.pipelineInterfaces[tidEmu].getCore().currentThreads++; // current
			// number of
			// threads in
			// this pipeline
			System.out.println("num of threads on core " + tidApp + " = "
					+ this.pipelineInterfaces[tidEmu].getCore().currentThreads);
			// INSTRUCTION_THRESHOLD = (numInstructionsInPool/currentEMUTHREADS)
			// - (EMUTHREADS-currentEMUTHREADS)*100;
		}
		if (pnew.ip == thread.pold.ip) {
			thread.packets.add(pnew);
		} else {

			int oldLength = inputToPipeline[tidEmu].size();

			long numHandledInsn = 0;
			int numMicroOpsBefore = thread.outstandingMicroOps.size();

			ObjParser
					.fuseInstruction(tidApp, thread.packets.get(0).ip, thread.packets, thread.outstandingMicroOps, tid);

			Statistics.setNumCISCInsn(Statistics.getNumCISCInsn(tid, tidEmu) + 1, tid, tidEmu);

			int numMicroOpsAfter = thread.outstandingMicroOps.size();
			if (numMicroOpsAfter > numMicroOpsBefore) {
				numHandledInsn = 1;
			} else {
				numHandledInsn = 0;
			}

			// System.out.println(" Size of new ins  " +(numMicroOpsAfter -
			// numMicroOpsBefore));
			// For one CISC instruction, we generate x micro-operations.
			// We set the CISC ip of the first micro-op to the original CISC ip.
			// IP of all the remaining micro-ops is set to -1(Invalid).
			// This ensures that we do not artificially increase the hit-rate of
			// instruction cache.
			for (int i = numMicroOpsBefore; i < numMicroOpsAfter; i++) {
				if (i == numMicroOpsBefore) {
					thread.outstandingMicroOps.peek(i).setCISCProgramCounter(thread.packets.get(0).ip);
				} else {
					thread.outstandingMicroOps.peek(i).setCISCProgramCounter(-1);
				}
			}

			// If I am running multiple benchmarks, the addresses of all the
			// benchmarks must
			// be tagged with benchmark ID. The tagging happens only if :
			// (a) there are multiple benchmarks (b) the benchmark id for this
			// thread is not zero
			if (Main.benchmarkThreadMapping.getNumBenchmarks() > 1 && tidApp > 0
					&& Main.benchmarkThreadMapping.getBenchmarkIDForThread(tidApp) != 0) {
				for (int i = numMicroOpsBefore; i < numMicroOpsAfter; i++) {
					thread.outstandingMicroOps.peek(i).changeAddressesForBenchmark(
							Main.benchmarkThreadMapping.getBenchmarkIDForThread(tidApp));
				}
			}
			
			// Either add all outstanding micro-ops or none.
			if (thread.outstandingMicroOps.size() < this.inputToPipeline[tidEmu].spaceLeft()) {
				// add outstanding micro-operations to input to pipeline
				while (thread.outstandingMicroOps.isEmpty() == false) {
					this.inputToPipeline[tidEmu].enqueue(thread.outstandingMicroOps.dequeue());
				}
			} else {
				isSpaceInPipelineBuffer = false;
			}

			int newLength = inputToPipeline[tidEmu].size();
			noOfMicroOps[tidEmu] += newLength - oldLength;
			

			Statistics.setNumHandledCISCInsn(Statistics.getNumHandledCISCInsn(tid, tidEmu) + numHandledInsn, tid,
					tidEmu);

			if (!thread.halted && (this.inputToPipeline[tidEmu].size() > Main.INSTRUCTION_THRESHOLD)) {
				thread.halted = true;
			}

			/*
			 * long temp = noOfMicroOps[tidEmu] % 1000000; if (temp < 1) {
			 * System.out.println("number of micro-ops = " +
			 * noOfMicroOps[tidEmu] + " on core " + tidApp); }
			 */

			thread.pold = pnew;
			thread.packets.clear();
			thread.packets.add(thread.pold);
		}
		return isSpaceInPipelineBuffer;
	}

	public void setThreadState(int tid, boolean cond) {
		threadParams[tid].halted = cond;
	}

	protected void signalFinish(int tidEmu) {
		this.inputToPipeline[tidEmu].enqueue(new Instruction(OperationType.inValid, null, null, null));
		int tidApp = IpcBase.getAppId(this.tid, tidEmu);
		IpcBase.glTable.getStateTable().get((Integer) tidApp).lastTimerseen = Long.MAX_VALUE;// (long)-1>>>1;
		// System.out.println(tidApp+" pin thread got -1");

		threadParams[tidEmu].finished = true;

		// hack, should ideally listen for fini function from pin. although this
		// will always work
		/*
		 * if (IpcBase.getAppId(tid, tidEmu) == 0) { for (int i = 0; i <
		 * SystemConfig.maxNumJavaThreads; i++) ipcType.termination[i] = true; }
		 */
	}
}
