package emulatorinterface.translator.qemuTranslationCache;

import main.Main;
import generic.Instruction;
import generic.InstructionList;

//Translation Cache for caching translations of qemu assembly instructions to micro-ops
public class TranslatedInstructionCache {
	private static final int cacheSize = 100000;
	private long cacheHit = 0;
	private long cacheMiss = 0;
	public LRUCache translatedInstructionTable;

	private void createTranslatedInstructionTable() {
		translatedInstructionTable = new LRUCache(cacheSize);
	}

	public void add(String asmText, InstructionList instructionList) {
		if (translatedInstructionTable == null) {
			createTranslatedInstructionTable();
		}
		InstructionList instList = null;
		for (int i = 0; i < instructionList.length(); i++) {
			if (instructionList.get(i).getOperationType() == null)
				continue;
			Instruction newInsn = Main.myPool[0].getInstructionPool().borrowObject();
			newInsn.copy(instructionList.get(i));
			if (instList == null)
				instList = new InstructionList(instructionList.length());
			instList.appendInstruction(newInsn);
		}
		if (instList != null)
			translatedInstructionTable.put(asmText, instList);
	}

	public InstructionList getInstructionList(String instText) {
		if (translatedInstructionTable == null) {
			return null;
		} else {
			InstructionList instructionList = translatedInstructionTable.get(instText);
			InstructionList instructionListToReturn = null;

			// increment references for each argument for each instruction
			for (int i = 0; i < instructionList.length(); i++) {
				if (instructionList.get(i).getOperationType() == null)
					continue;
				if (instructionListToReturn == null)
					instructionListToReturn = new InstructionList(instructionList.getListSize());
				Instruction newInstruction = new Instruction();
				newInstruction.copy(instructionList.get(i));
				instructionListToReturn.appendInstruction(newInstruction);
			}
			return instructionListToReturn;
		}
	}

	public boolean isPresent(String instText) {
		if (translatedInstructionTable == null) {
			return false;
		} else {
			boolean ret = translatedInstructionTable.containsKey(instText);
			if (ret) {
				cacheHit++;
			} else {
				cacheMiss++;
			}

			return ret;
		}
	}

	public float getHitRate() {
		if ((cacheHit + cacheMiss) == 0) {
			return -1f;
		} else {
			return ((float) (cacheHit) / (float) (cacheHit + cacheMiss));
		}
	}
}
