/*****************************************************************************
				Tejas Simulator
------------------------------------------------------------------------------------------------------------

   Copyright [2010] [Indian Institute of Technology, Delhi]
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
------------------------------------------------------------------------------------------------------------

	Contributors:  Prathmesh Kallurkar
 *****************************************************************************/

package emulatorinterface.translator.x86.instruction;

import emulatorinterface.translator.InvalidInstructionException;
import emulatorinterface.translator.x86.registers.Registers;
import emulatorinterface.translator.x86.registers.TempRegisterNum;
import generic.Instruction;
import generic.Operand;
import generic.InstructionList;

public class ReturnOp implements X86StaticInstructionHandler 
{
	public void handle(long instructionPointer, 
			Operand operand1, Operand operand2, Operand operand3,
			InstructionList instructionArrayList,
			TempRegisterNum tempRegisterNum, int tid)
					throws InvalidInstructionException
	{
		if ((operand1 == null || operand1.isImmediateOperand())
				&& operand2 == null && operand3 == null) 
		{
      // Create stack-pointer and [stack-pointer]
		  Operand stackPointer = Registers.getStackPointer();
		  Operand stackPointerLocation = Operand.getMemoryOperand(stackPointer, null);

			Operand newInstructionPointer;
			newInstructionPointer = Registers.getInstructionPointer();

			// load the new instruction-pointer from the stack
			instructionArrayList.appendInstruction(Instruction.getLoadInstruction(stackPointerLocation,	newInstructionPointer, tid));

			// perform an unconditional jump to the new location
			(new UnconditionalJump()).handle(instructionPointer,
					newInstructionPointer, null, null, instructionArrayList, tempRegisterNum, tid);
      
      if(operand1==null) {
        
			  //stack-pointer = stack-pointer - 4
			  instructionArrayList.appendInstruction(Instruction.getIntALUInstruction(stackPointer, Operand.getImmediateOperand(), stackPointer, tid));

      } else if (operand1 != null && operand1.isImmediateOperand()) {
        //stack-pointer = stack-pointer - (operand1+4)
        
        Operand tempRegister = Registers.getTempIntReg(tempRegisterNum);
        
        //temp = operand1 + 4
			  instructionArrayList.appendInstruction(Instruction.getIntALUInstruction(operand1, Operand.getImmediateOperand(), tempRegister, tid));

        // stack-pointer = stack-pointer + temp-Register
			  instructionArrayList.appendInstruction(Instruction.getIntALUInstruction(stackPointer, tempRegister, stackPointer, tid));
			} else {
      	
        misc.Error.invalidOperation("Return Operation", operand1, operand2,
					operand3);
      }
		}
		else 
		{
			misc.Error.invalidOperation("Return Operation", operand1, operand2,
					operand3);
		}
	}
}
