package generic;

import java.util.concurrent.PriorityBlockingQueue;

public class EventQueue 
{
	private Core core;
	private ClockDomain clockDomain;
	static final int queueSize = 1024;
	private PriorityBlockingQueue<Event> priorityQueue;
	
	public EventQueue(ClockDomain clockDomain, Core core) 
	{
		priorityQueue = new PriorityBlockingQueue<Event>(1, new EventComparator());
		this.clockDomain = clockDomain;
		this.core = core;
	}

	public PriorityBlockingQueue<Event> getPriorityQueue()
	{
		return priorityQueue;
	}
	
	
	public void addEvent(Event event)
	{
		priorityQueue.add(event);
	}
	
	public void processEvents()
	{
		
		Event event;
		long eventTime;
		
		long currentClockTime = clockDomain.currentTime;
		while(!priorityQueue.isEmpty())
		{
			//get the eventTime of the event on the head of the queue.
			Event e = priorityQueue.peek();
			eventTime = e.getEventTime();
			if (eventTime <= currentClockTime)
			{
				event = priorityQueue.remove();
				//System.out.println(threadId + ": RS1");
				event.getProcessingElement().handleEvent(this, event);
				//System.out.println(threadId + ": RS2");
			}
			else
			{
				break;
			}
		}
		
	}
	
	public Core getCore() {
		return core;
	}
	public ClockDomain getClockDomain() {
		return clockDomain;
	}
}