package main;

import config.EmulatorConfig;
import config.EmulatorType;
import emulatorinterface.communication.CustomAsmCharPool;
import generic.CustomInstructionPool;

public class CustomObjectPool {
	
	private CustomInstructionPool instructionPool;
	private CustomAsmCharPool customAsmCharPool;
	
	public void initCustomPools(int maxApplicationThreads, int staticInstructionPoolSize) {
		
		// Create Pools of Instructions, Operands and AddressCarryingEvents
		int runTimePoolPerAppThread =  Main.INSTRUCTION_THRESHOLD;
		int staticTimePool = staticInstructionPoolSize;
		
		// best case -> single threaded application
		int minInstructionPoolSize = staticTimePool + runTimePoolPerAppThread;
		int maxInstructionPoolSize = staticTimePool + runTimePoolPerAppThread * maxApplicationThreads * 2;
				
		/* custom pool */
		System.out.println("creating instruction pool..");
		setInstructionPool(new CustomInstructionPool(minInstructionPoolSize, maxInstructionPoolSize));
		
		if(EmulatorConfig.emulatorType==EmulatorType.none) {
			System.out.println("creating custom asm-char pool. max threads = " + maxApplicationThreads);
			customAsmCharPool = new CustomAsmCharPool(maxApplicationThreads);
		}
	}

	public CustomInstructionPool getInstructionPool() {
		return instructionPool;
	}

	public void setInstructionPool(CustomInstructionPool instructionPool) {
		this.instructionPool = instructionPool;
	}

	public CustomAsmCharPool getCustomAsmCharPool() {
		return customAsmCharPool;
	}

	public void setCustomAsmCharPool(CustomAsmCharPool customAsmCharPool) {
		this.customAsmCharPool = customAsmCharPool;
	}
}
