package pipeline.multi_issue_inorder;

import config.SimulationConfig;
import config.SystemConfig;
import generic.Core;
import generic.Event;
import generic.EventQueue;
import generic.GenericCircularQueue;
import generic.Instruction;
import generic.OperationType;
import generic.PortType;
import generic.SimulationElement;

public class FetchUnitIn_MII extends SimulationElement {
	Core core;
	MultiIssueInorderExecutionEngine containingExecutionEngine;

	public GenericCircularQueue<Instruction> inputToPipeline;
	StageLatch_MII ifId_latch;

	Instruction fetchBuffer[];
	public int fetchBufferCapacity;
	private int fetchFillCount; // Number of instructions in the fetch buffer
	private int fetchBufferIndex; // Index to first instruction to be popped out
									// of fetch buffer
	private boolean fetchBufferStatus[]; // To check whether request to ICache
											// is complete or not

	private boolean sleep; // The boolean to stall the pipeline when a sync
							// request is received
	int syncCount;
	long numRequestsSent;
	int numRequestsAcknowledged;

	long instCtr; // for debug

	public FetchUnitIn_MII(Core core, EventQueue eventQueue,
			MultiIssueInorderExecutionEngine execEngine) {
		super(PortType.Unlimited, -1, -1, -1, -1);

		this.core = core;
		this.containingExecutionEngine = execEngine;

		this.ifId_latch = execEngine.getIfIdLatch();

		this.fetchBufferCapacity = (int) (core.getIssueWidth() * (SystemConfig.core[core
				.getCore_number()].getICacheLatency()));
		this.fetchBuffer = new Instruction[this.fetchBufferCapacity];
		this.fetchFillCount = 0;
		this.fetchBufferIndex = 0;
		this.fetchBufferStatus = new boolean[this.fetchBufferCapacity];
		for (int i = 0; i < this.fetchBufferCapacity; i++) {
			this.fetchBufferStatus[i] = false;
		}

		this.sleep = false;
		this.syncCount = 0;
		this.numRequestsSent = 0;
		this.numRequestsAcknowledged = 0;
		instCtr = 0;
	}

	public void fillFetchBuffer(MultiIssueInorderPipeline inorderPipeline) {
		if (inputToPipeline.isEmpty())
			return;
		
		/*boolean checkTranslatorSpeed = true;

		if (checkTranslatorSpeed) {
			Instruction inst;
			while ((inst = inputToPipeline.dequeue()) != null) {
				if (inst.getOperationType() == OperationType.inValid) {
					containingExecutionEngine.setExecutionComplete(true);
				}
				//CustomObjectPool.getInstructionPool().returnObject(inst);
			}

			return;
		}*/
		
		Instruction newInstruction = null;
		for (int i = (this.fetchBufferIndex + this.fetchFillCount)
				% this.fetchBufferCapacity; this.fetchFillCount < this.fetchBufferCapacity; i = (i + 1)
				% this.fetchBufferCapacity) {

			if (containingExecutionEngine.multiIssueInorderCoreMemorySystem
					.getiCache().isBusy()) {
				break;
			}

			newInstruction = inputToPipeline.pollFirst();

			if (newInstruction == null)
				return;
			numRequestsSent++;
			if (newInstruction.getOperationType() == OperationType.inValid) {
				this.fetchBuffer[i] = newInstruction;
				this.fetchBufferStatus[i] = true;
				this.fetchFillCount++;
			}

			else {
				this.fetchBuffer[i] = newInstruction;
				this.fetchFillCount++;
				if (newInstruction.getOperationType() != OperationType.sync) {
					newInstruction.setSerialNo(instCtr++);
				}

				// The first micro-operation of an instruction has a valid CISC
				// IP. All the subsequent
				// micro-ops will have IP = -1(meaning invalid). We must not
				// forward this requests to iCache.
				if (SimulationConfig.detachMemSysInsn
						|| newInstruction.getCISCProgramCounter() == -1) {
					this.fetchBufferStatus[i] = true;
				} else {
					this.fetchBufferStatus[i] = false;
					containingExecutionEngine.multiIssueInorderCoreMemorySystem
							.issueRequestToInstrCache(newInstruction
									.getCISCProgramCounter());
				}
			}
		}
	}

	public void performFetch(MultiIssueInorderPipeline inorderPipeline) {
		if (containingExecutionEngine.getMispredStall() > 0) {
			containingExecutionEngine.decrementMispredStall(1);
			return;
		}

		Instruction ins;

		if (!this.fetchBufferStatus[this.fetchBufferIndex])
			containingExecutionEngine.incrementInstructionMemStall(1);

		// move to the IF-ID latch those instructions that have completed
		// fetch from the i-cache
		while (!this.sleep && this.fetchFillCount > 0
				&& this.fetchBufferStatus[this.fetchBufferIndex]
				&& this.ifId_latch.isFull() == false) {
			ins = this.fetchBuffer[this.fetchBufferIndex];
			
			//System.out.println("Fetch " + core.getCore_number() + "      " + ins);
			this.ifId_latch.add(ins, core.getClockDomain().getCurrentTime() + 1);
			this.fetchFillCount--;
			this.fetchBuffer[this.fetchBufferIndex] = null;
			this.fetchBufferStatus[this.fetchBufferIndex] = false;
			this.fetchBufferIndex = (this.fetchBufferIndex + 1)
					% this.fetchBufferCapacity;

			/*if (SimulationConfig.debugMode core.getCore_number() == 0 && ins.getOperationType() == OperationType.load) {
				System.out.println("fetched : " + core.getClockDomain().getCurrentTime()
						/ core.getStepSize() + ins + "\n");
			}*/
		}

		// remove micro-ops from inputToPipeline and issue request to the
		// i-cache
		fillFetchBuffer(inorderPipeline);
	}

	public void setSleep(boolean _sleep) {
		this.sleep = _sleep;
	}

	public boolean getSleep() {
		return this.sleep;
	}

	public void sleepThePipeline() {
		System.out.println("sleeping pipeline" + this.core.getCore_number()
				+ "...!!");
		this.syncCount--;
		this.sleep = true;
	}

	public GenericCircularQueue<Instruction> getInputToPipeline() {
		return this.inputToPipeline;
	}

	public void setInputToPipeline(GenericCircularQueue<Instruction> inpList) {
		this.inputToPipeline = inpList;
	}

	public void resumePipeline() {
		System.out.println("Resuming the pipeline "
				+ this.core.getCore_number() + "...!!");
		this.syncCount++;
		this.sleep = false;
	}

	@Override
	public void handleEvent(EventQueue eventQ, Event event) {

	}

	public void processCompletionOfMemRequest(long requestedAddress) {
		for (int i = 0; i < this.fetchBufferCapacity; i++) {
			if (this.fetchBuffer[i] != null
					&& this.fetchBuffer[i].getCISCProgramCounter() == requestedAddress
					&& this.fetchBufferStatus[i] == false) {
				this.fetchBufferStatus[i] = true;
			}
		}
	}

}
