package pipeline.multi_issue_inorder;

import config.SimulationConfig;
import memorysystem.AddressCarryingEvent;
import generic.Core;
import generic.Event;
import generic.EventQueue;
import generic.Instruction;
import generic.Operand;
import generic.OperationType;
import generic.PortType;
import generic.RequestType;
import generic.SimulationElement;

public class MemUnitIn_MII extends SimulationElement{
	
	Core core;
	MultiIssueInorderExecutionEngine containingExecutionEngine;
	EventQueue eventQueue;
	StageLatch_MII exMemLatch;
	StageLatch_MII memWbLatch;
	
	long instCtr; //for debug

	public MemUnitIn_MII(Core core, MultiIssueInorderExecutionEngine execEngine) {
		super(PortType.Unlimited, -1, -1 , -1, -1);
		this.core = core;
		this.eventQueue = core.getEventQueue();
		containingExecutionEngine = execEngine;
		exMemLatch = execEngine.getExMemLatch();
		memWbLatch = execEngine.getMemWbLatch();
		
		instCtr = 0;
	}
	
	public void performMemEvent(MultiIssueInorderPipeline inorderPipeline)
	{
		if(containingExecutionEngine.getMispredStall() > 0)
		{
			return;
		}
		
		Instruction ins = null;
		
		while(exMemLatch.isEmpty() == false
				&& memWbLatch.isFull() == false)
		{
			ins = exMemLatch.peek(0, core.getClockDomain().getCurrentTime());
			
			if(ins != null)
			{
				long lat = 1;
				if(ins.getOperationType() == OperationType.load)
				{
					//issue load request
					if(!SimulationConfig.detachMemSysData)
					{		
						boolean memReqIssued = containingExecutionEngine.multiIssueInorderCoreMemorySystem.issueRequestToL1Cache(
								RequestType.Cache_Read,
								ins.getSourceOperand1MemValue());
						
						
						if(memReqIssued == false)
						{
							break;
						}
						lat = Long.MAX_VALUE - core.getClockDomain().getCurrentTime();
						//System.out.println("REQUEST ISSUED : " + core.getClockDomain().getCurrentTime()/core.getStepSize() + ins + "\n");
					}
					
					//set instruction's MEM stage completion time to Long.MAX_VALUE
					lat = 1;
					//TODO: Confirm this with raj sir
				}
				if(ins.getOperationType() == OperationType.store)
				{
					if(!SimulationConfig.detachMemSysData)
					{
						boolean memReqIssued = containingExecutionEngine.multiIssueInorderCoreMemorySystem.issueRequestToL1Cache(
								RequestType.Cache_Write,
								ins.getSourceOperand1MemValue());
						
						if(memReqIssued == false)
						{
							break;
						}
					}
				}
				
//				if(ins.getSerialNo() != instCtr && ins.getOperationType() != OperationType.inValid)
//				{
//					misc.Error.showErrorAndExit("mem out of order!!");
//				}
				instCtr++;
				
				//move ins to next stage
				memWbLatch.add(ins, core.getClockDomain().getCurrentTime() + lat);
				exMemLatch.poll(core.getClockDomain().getCurrentTime());
				
			/*	if (SimulationConfig.debugMode core.getCore_number() == 0 && ins.getOperationType() == OperationType.load)  
				{
					System.out.println("MEM : " + core.getClockDomain().getCurrentTime()/core.getStepSize() + ins + "\n");
				}*/
			}
			else
			{
				break;
			}
		}
	}


	@Override
	public void handleEvent(EventQueue eventQ, Event event) 
	{
		updateInstructionCompletions(((AddressCarryingEvent)event).getAddress());
	}
	
	public void processCompletionOfMemRequest(long requestedAddress)
	{
		updateInstructionCompletions(requestedAddress);
		containingExecutionEngine.noOfOutstandingLoads--;
	}
	
	void updateInstructionCompletions(long address)
	{
		Instruction instructions[] = memWbLatch.getInstructions();
		long instructionCompletesAt[] = memWbLatch.getInstructionCompletesAt();
		
		for(int i = 0; i < containingExecutionEngine.getIssueWidth(); i++)
		{
			if(instructions[i] != null
					&& instructions[i].getOperationType() == OperationType.load
					&& instructions[i].getSourceOperand1MemValue() == address
					&& instructionCompletesAt[i] > core.getClockDomain().getCurrentTime())
			{
				instructionCompletesAt[i] = core.getClockDomain().getCurrentTime();
				
				Operand destOpnd = instructions[i].getDestinationOperand();
				if(destOpnd.isIntegerRegisterOperand())
				{
					containingExecutionEngine.getValueReadyInteger()[(int)destOpnd.getValue()]
							 												= core.getClockDomain().getCurrentTime();
				}
				else if(destOpnd.isFloatRegisterOperand())
				{
					containingExecutionEngine.getValueReadyFloat()[(int)destOpnd.getValue()]
							 												= core.getClockDomain().getCurrentTime();
				}
			}
		}
	}
}
