package pipeline.multi_issue_inorder;

import java.io.FileWriter;
import java.io.IOException;

import generic.Instruction;
import config.EnergyConfig;
import config.SimulationConfig;
import main.Main;
import generic.Core;
import generic.Event;
import generic.EventQueue;
import generic.Operand;
import generic.OperationType;
import generic.PinPointsProcessing;
import generic.PortType;
import generic.SimulationElement;

public class WriteBackUnitIn_MII extends SimulationElement{
	
	Core core;
	MultiIssueInorderExecutionEngine containingExecutionEngine;
	StageLatch_MII memWbLatch;
	
	long instCtr; //for debug
	
	long numIntRegFileAccesses;
	long numFloatRegFileAccesses;
	
	public WriteBackUnitIn_MII(Core core, MultiIssueInorderExecutionEngine execEngine)
	{
		super(PortType.Unlimited, -1, -1 , -1, -1);
		this.core = core;
		containingExecutionEngine = execEngine;
		memWbLatch = execEngine.getMemWbLatch();
		
		instCtr = 0;
	}
	
	public void performWriteBack(MultiIssueInorderPipeline inorderPipeline)
	{
		if(containingExecutionEngine.getMispredStall() > 0)
		{
			return;
		}
		
		Instruction ins = null;
		
		while(memWbLatch.isEmpty() == false)
		{
			ins = memWbLatch.peek(0, core.getClockDomain().getCurrentTime());
			if(ins != null)
			{
				//check if simulation complete
				if(ins.getOperationType()==OperationType.inValid)
				{
					this.core.currentThreads--;
					
					if(containingExecutionEngine.isExecutionComplete() == false && this.core.currentThreads <= 0){   //set exec complete only if there are no other thread already 
														  //assigned to this pipeline	
						containingExecutionEngine.setExecutionComplete(true);
						if(SimulationConfig.pinpointsSimulation == false)
						{					
							containingExecutionEngine.setTimingStatistics();			
							containingExecutionEngine.setPerCoreMemorySystemStatistics();
						}
						else
						{
							PinPointsProcessing.processEndOfSlice();
						}
					}
				}
				else
				{
					if(core.getNoOfInstructionsExecuted()%1000000==0)
					{
						System.out.println(core.getNoOfInstructionsExecuted()/1000000 + " million done" + " by core "+core.getCore_number() 
								+ " global clock cycle " + core.getClockDomain().getCurrentTime());
					}
					core.incrementNoOfInstructionsExecuted();
				}
				
				//increment register file accesses for power statistics
				
				//operand fetch
				incrementNumRegFileAccesses(ins.getSourceOperand1(), 1);
				incrementNumRegFileAccesses(ins.getSourceOperand2(), 1);
				
				//write-back
				incrementNumRegFileAccesses(ins.getDestinationOperand(), 1);
				if(ins.getOperationType() == OperationType.xchg)
				{
					incrementNumRegFileAccesses(ins.getSourceOperand1(), 1);
					if(ins.getSourceOperand1().getValue() != ins.getSourceOperand2().getValue()
							|| ins.getSourceOperand1().getOperandType() != ins.getSourceOperand2().getOperandType())
					{
						incrementNumRegFileAccesses(ins.getSourceOperand2(), 1);
					}
				}
				
//				if(ins.getSerialNo() != instCtr && ins.getOperationType() != OperationType.inValid)
//				{
//					misc.Error.showErrorAndExit("wb out of order!!");
//				}
				instCtr++;	
				
			/*	if (SimulationConfig.debugMode core.getCore_number() == 0 && ins.getOperationType() == OperationType.load) 
				{
					System.out.println("write back : " + core.getClockDomain().getCurrentTime()/core.getStepSize() +  ins + "\n");
				}
			*/	
				memWbLatch.poll(core.getClockDomain().getCurrentTime());				
				try
				{
					Main.myPool[core.getCore_number()].getInstructionPool().returnObject(ins);
					core.numReturns++;
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
			else
			{
				break;
			}
		}
	}

	@Override
	public void handleEvent(EventQueue eventQ, Event event) {

	}
	
	void incrementNumRegFileAccesses(Operand operand, int incrementBy)
	{
		if(operand == null)
		{
			return;
		}
		
		if(operand.isIntegerRegisterOperand())
		{
			incrementNumIntRegFileAccesses(incrementBy);
		}
		else if(operand.isFloatRegisterOperand())
		{
			incrementNumFloatRegFileAccesses(incrementBy);
		}
	}
	
	void incrementNumIntRegFileAccesses(int incrementBy)
	{
		numIntRegFileAccesses += incrementBy;
	}
	
	void incrementNumFloatRegFileAccesses(int incrementBy)
	{
		numFloatRegFileAccesses += incrementBy;
	}
	
	public EnergyConfig calculateAndPrintEnergy(FileWriter outputFileWriter, String componentName) throws IOException
	{
		EnergyConfig totalPower = new EnergyConfig(0, 0);
		EnergyConfig intRegFilePower = new EnergyConfig(core.getIntRegFilePower(), numIntRegFileAccesses);
		totalPower.add(totalPower, intRegFilePower);
		EnergyConfig floatRegFilePower = new EnergyConfig(core.getFpRegFilePower(), numFloatRegFileAccesses);
		totalPower.add(totalPower, floatRegFilePower);
		
		intRegFilePower.printEnergyStats(outputFileWriter, componentName + ".int");
		floatRegFilePower.printEnergyStats(outputFileWriter, componentName + ".float");
		
		return totalPower;
	}
}
